
/**
 * A PDDL generator that generates PDDL descriptions of domains and problems of
 * No-Mystery.
 *
 * @author Dietrich Zerr
 */
public class NoMystery implements Problem {
    
    /**
     * Number of nodes in the solvable part of the problem.
     */
    int numberOfNodesSolvable = 2;
    /**
     * Number of nodes in the unsolvable part of the problem.
     */
    int numberOfNodesUnsolvable = 2;
    /**
     * Whether the two graphs/part-problems are connected.
     */
    boolean graphsConnected = false;
    /**
     * Whether the nodes in the unsolvable part are all connected.
     */
    boolean allUnsolvableNodesConnected = false;
    /**
     * Number of packages in the solvable part.
     */
    int numberOfPackagesSolvable = 3;
    /**
     * Number of packages in the unsolvable part.
     */
    int numberOfPackagesUnsolvable = 3;

    /**
     * The domain of a No-Mystery instance can be the same for all problems.
     * It's a static domain PDDL taken from the Fast Downward benchmarks folder.
     *
     * @param properties - not used
     * @return a Sokoban PDDL domain description
     */
    @Override
    public String getPDDLDomain(int[] properties) throws IllegalArgumentException {
        if (properties.length == 0) {
            throw new IllegalArgumentException("Please specify at least the unsolvability mode for No-Mystery.\n"
                    + "The No-Mystery call should look like this:\n"
                    + "Problemgenerator [<#problems to generate>] NoMystery <mode> [<#nodes in unsolvable part>] [<#nodes in solvable part>] [<#packages in unsolvable part>] [<#packages in solvable part>] [<graph connection status]>\n"
                    + "Please see the readme for more details.");
        }

        StringBuilder pddl = new StringBuilder(2000);
        pddl.append("(define (domain no-mystery)\n");
        pddl.append("   (:predicates\n");
        pddl.append("       (fuel-number ?x)\n");
        pddl.append("       (capacity-number ?x)\n");
        pddl.append("       (location ?x)\n");
        pddl.append("       (vehicle ?x)\n");
        pddl.append("       (package ?x)\n");
        pddl.append("       (connected ?n1 ?n2)\n");
        pddl.append("       (at ?v ?n)\n");
        pddl.append("       (in ?c ?v)\n");
        pddl.append("       (fuel ?n ?a)\n");
        pddl.append("       (capacity ?v ?s)\n");
        pddl.append("       (fuel-predecessor ?i ?j)\n");
        pddl.append("       (capacity-predecessor ?i ?j))\n");
        pddl.append("\n");
        pddl.append("   (:action load\n");
        pddl.append("       :parameters (?c ?v ?n ?s1 ?s2)\n");
        pddl.append("       :precondition (and (package ?c)\n");
        pddl.append("                          (vehicle ?v)\n");
        pddl.append("                          (at ?c ?n)\n");
        pddl.append("                          (at ?v ?n)\n");
        pddl.append("                          (location ?n)\n");
        pddl.append("                          (capacity ?v ?s2)\n");
        pddl.append("                          (capacity-number ?s2)\n");
        pddl.append("                          (capacity-predecessor ?s1 ?s2)\n");
        pddl.append("                          (capacity-number ?s1))\n");
        pddl.append("       :effect (and (not (at ?c ?n))\n");
        pddl.append("                    (in ?c ?v)\n");
        pddl.append("                    (not (capacity ?v ?s2))\n");
        pddl.append("                    (capacity ?v ?s1)))\n");
        pddl.append("   (:action drive\n");
        pddl.append("       :parameters (?v ?n1 ?n2 ?l1 ?l2)\n");
        pddl.append("       :precondition (and (at ?v ?n1)\n");
        pddl.append("                          (location ?n1)\n");
        pddl.append("                          (vehicle ?v)\n");
        pddl.append("                          (connected ?n1 ?n2)\n");
        pddl.append("                          (location ?n2)\n");
        pddl.append("                          (fuel ?n1 ?l2)\n");
        pddl.append("                          (fuel-predecessor ?l1 ?l2))\n");
        pddl.append("       :effect (and (not (at ?v ?n1))\n");
        pddl.append("                         (at ?v ?n2)\n");
        pddl.append("                         (not (fuel ?n1 ?l2))\n");
        pddl.append("                         (fuel ?n1 ?l1)))\n");
        pddl.append("   (:action unload\n");
        pddl.append("       :parameters (?c ?v ?n ?s1 ?s2)\n");
        pddl.append("       :precondition (and (in ?c ?v)\n");
        pddl.append("                          (package ?c)\n");
        pddl.append("                          (vehicle ?v)\n");
        pddl.append("                          (at ?v ?n)\n");
        pddl.append("                          (location ?n)\n");
        pddl.append("                          (capacity ?v ?s1)\n");
        pddl.append("                          (capacity-predecessor ?s1 ?s2))\n");
        pddl.append("       :effect (and (not (in ?c ?v))\n");
        pddl.append("                    (at ?c ?n)\n");
        pddl.append("                    (not (capacity ?v ?s1))\n");
        pddl.append("                    (capacity ?v ?s2))))");
        return pddl.toString();
    }

    /**
     * A No-Mystery problem depends on the attributes. < br >
     * attributes array: < br >
     * index - attribute < br >
     * 0 - < br >
     * 1 - < br >
     * 2 - < br >
     * 3 - < br >
     *
     * @param properties No-Mystery problem properties
     * @param instanceNumber the number of the problem instance
     * @return a Sokoban PDDL problem description
     */
    @Override
    public String getPDDLProblem(int[] properties, int instanceNumber) throws IllegalArgumentException {
        if (properties.length >= 1) {
            if (properties[0] >= 2) {
                numberOfNodesUnsolvable = properties[0];
                numberOfPackagesUnsolvable = properties[0];
            } else if (properties[0] == 1) {
                numberOfNodesUnsolvable = 2;
                numberOfPackagesUnsolvable = 2;
                if (instanceNumber == 1) {
                    System.out.println("Number of nodes must be at least 2. Number of nodes in the unsolvable part has been set to default " + numberOfNodesUnsolvable + ".");
                }
            } else {
                if (instanceNumber == 1) {
                    System.out.println("Number of nodes cannot be negative. Number of nodes in the unsolvable part has been set to default (" + numberOfNodesUnsolvable + ").");
                }
            }
        }
        if (properties.length >= 2) {
            if (properties[1] >= 2) {
                numberOfNodesSolvable = properties[1];
                numberOfPackagesSolvable = properties[1];
            } else if (properties[1] == 1) {
                numberOfNodesSolvable = 2;
                numberOfPackagesSolvable = 2;
                if (instanceNumber == 1) {
                    System.out.println("Number of nodes must be at least 2. Number of nodes in the solvable part has been set to " + numberOfNodesSolvable + ".");
                }
            } else {
                if (instanceNumber == 1) {
                    System.out.println("Number of nodes cannot be negative. Number of nodes in the solvable part has been set to default (" + numberOfNodesSolvable + ").");
                }
            }
        }
        if (properties.length >= 3) {
            if (properties[2] >= 0) {
                numberOfPackagesUnsolvable = properties[2];
            } else {
                numberOfPackagesUnsolvable = 0;
                if (instanceNumber == 1) {
                    System.out.println("Number of packages cannot be negative. Number of packages in the unsolvable part has been set to " + numberOfPackagesUnsolvable + ".");
                }
            }
        }
        if (properties.length >= 4) {
            if (properties[3] >= 0) {
                numberOfPackagesSolvable = properties[3];
            } else {
                numberOfPackagesSolvable = 0;
                if (instanceNumber == 1) {
                    System.out.println("Number of packages cannot be negative. Number of packages in the solvable part has been set to " + numberOfPackagesSolvable + ".");
                }
            }
        }
        if (properties.length >= 5) {
            if (properties[4] == 0) {
                graphsConnected = false;
            } else if (properties[4] == 1) {
                graphsConnected = true;
            } else {
                if (instanceNumber == 1) {
                    System.out.println("Graph connection status can only be 0 (false) or 1 (true). Graph connection status has been set to default (" + graphsConnected + ").");
                }
            }
        }
        if (properties.length >= 6) {
            if (properties[5] == 0) {
                allUnsolvableNodesConnected = false;
            } else if (properties[5] == 1) {
                allUnsolvableNodesConnected = true;
            } else {
                if (instanceNumber == 1) {
                    System.out.println("Unsolvable nodes connection status can only be 0 (false) or 1 (true). Unsolvable nodes connection status has been set to default (" + allUnsolvableNodesConnected + ").");
                }
            }
        }

        StringBuilder pddl = new StringBuilder(2000); // capacity estimated
        pddl.append(";; Nodes in the unsolvable subgraph: " + numberOfNodesUnsolvable + "\n");
        pddl.append(";; Nodes in the solvable subgraph: " + numberOfNodesSolvable + "\n");
        pddl.append(";; Packages in the unsolvable subgraph: " + numberOfPackagesUnsolvable + "\n");
        pddl.append(";; Packages in the solvable subgraph: " + numberOfPackagesSolvable + "\n");
        pddl.append(";; Solvable and unsolvable subgraphs are ");
        pddl.append(graphsConnected ? "connected.\n" : "not connected.\n");
        pddl.append(";; Unsolvable subgraph ");
        pddl.append(allUnsolvableNodesConnected ? "is a clique.\n" : "is not a clique.\n");
        pddl.append("\n");
        pddl.append("(define (problem no-mystery-problem)\n");
        pddl.append("  (:domain no-mystery)\n");
        pddl.append("  (:objects\n");
        pddl.append("    truck-1\n");
        pddl.append("    truck-2\n");
        pddl.append("    capacity-0\n");
        pddl.append("    capacity-1\n");
        for (int nodes = 1; nodes <= numberOfNodesUnsolvable; nodes++) {
            pddl.append("    uNode-" + nodes + "\n");
        }
        for (int nodes = 1; nodes <= numberOfNodesSolvable; nodes++) {
            pddl.append("    sNode-" + nodes + "\n");
        }
        for (int pack = 1; pack <= numberOfPackagesUnsolvable; pack++) {
            pddl.append("    uPackage-" + pack + "\n");
        }
        for (int pack = 1; pack <= numberOfPackagesSolvable; pack++) {
            pddl.append("    sPackage-" + pack + "\n");
        }
        int largerNumberOfPackages = (numberOfPackagesSolvable > numberOfPackagesUnsolvable) ? numberOfPackagesSolvable : numberOfPackagesUnsolvable;
        for (int pack = 0; pack <= largerNumberOfPackages; pack++) {
            pddl.append("    fuel-" + pack + "\n");
        }
        pddl.append("  )\n"); // objects close
        pddl.append("  (:init\n");
        pddl.append("    (vehicle truck-1)\n");
        pddl.append("    (at truck-1 uNode-1)\n");
        pddl.append("    (vehicle truck-2)\n");
        pddl.append("    (at truck-2 sNode-1)\n");
        pddl.append("    (capacity-number capacity-0)\n");
        pddl.append("    (capacity-number capacity-1)\n");
        pddl.append("    (capacity-predecessor capacity-0 capacity-1)\n");
        pddl.append("    (capacity truck-1 capacity-1)\n");
        pddl.append("    (capacity truck-2 capacity-1)\n");
        for (int nodes = 1; nodes <= numberOfNodesUnsolvable; nodes++) {
            pddl.append("    (location uNode-" + nodes + ")\n");
        }
        for (int nodes = 1; nodes <= numberOfNodesSolvable; nodes++) {
            pddl.append("    (location sNode-" + nodes + ")\n");
        }
        for (int pack = 1; pack <= numberOfPackagesUnsolvable; pack++) {
            pddl.append("    (package uPackage-" + pack + ")\n");
        }
        for (int pack = 1; pack <= numberOfPackagesSolvable; pack++) {
            pddl.append("    (package sPackage-" + pack + ")\n");
        }
        for (int pack = 0; pack <= largerNumberOfPackages; pack++) {
            pddl.append("    (fuel-number fuel-" + pack + ")\n");
        }
        for (int pack = 0; pack < largerNumberOfPackages; pack++) {
            pddl.append("    (fuel-predecessor fuel-" + pack + " fuel-" + (pack + 1) + ")\n");
        }
        for (int nodes = 1; nodes <= numberOfNodesUnsolvable; nodes++) {
            pddl.append("    (fuel uNode-" + nodes + " fuel-" + (numberOfPackagesUnsolvable - 1) + ")\n");
        }
        for (int nodes = 1; nodes <= numberOfNodesSolvable; nodes++) {
            pddl.append("    (fuel sNode-" + nodes + " fuel-" + numberOfPackagesSolvable + ")\n");
        }
        if (graphsConnected) {
            pddl.append("    (connected uNode-1 sNode-1)\n");
            pddl.append("    (connected sNode-1 uNode-1)\n");
        }
        for (int node1 = 1; node1 <= numberOfNodesSolvable; node1++) {
            for (int node2 = 1; node2 <= numberOfNodesSolvable; node2++) {
                if (node1 != node2) {
                    pddl.append("    (connected sNode-" + node1 + " sNode-" + node2 + ")\n");
                }
            }
        }

        StringBuilder uPackageStartNodes;
        StringBuilder uPackageGoalNodes;
        if (allUnsolvableNodesConnected) {
            for (int node1 = 1; node1 <= numberOfNodesUnsolvable; node1++) {
                for (int node2 = 1; node2 <= numberOfNodesUnsolvable; node2++) {
                    if (node1 != node2) {
                        pddl.append("    (connected uNode-" + node1 + " uNode-" + node2 + ")\n");
                    }
                }
            }
            uPackageStartNodes = new StringBuilder(30 * numberOfPackagesUnsolvable); // capacity estimated
            uPackageGoalNodes = new StringBuilder(33 * numberOfPackagesUnsolvable); // capacity estimated
            for (int pack = 1; pack <= numberOfPackagesUnsolvable; pack++) {
                uPackageStartNodes.append("    (at uPackage-" + pack + " uNode-1)\n");
                uPackageGoalNodes.append("      (at uPackage-" + pack + " uNode-" + (((pack - 1) % (numberOfNodesUnsolvable - 1)) + 2) + ")\n");
            }
        } else {
            for (int node = 2; node <= numberOfNodesUnsolvable; node++) {
                pddl.append("    (connected uNode-1 uNode-" + node + ")\n");
                pddl.append("    (connected uNode-" + node + " uNode-1)\n");
            }
            uPackageStartNodes = new StringBuilder(30 * numberOfPackagesUnsolvable); // capacity estimated
            uPackageGoalNodes = new StringBuilder(33 * numberOfPackagesUnsolvable); // capacity estimated
            for (int pack = 1; pack <= numberOfPackagesUnsolvable; pack++) {
                uPackageStartNodes.append("    (at uPackage-" + pack + " uNode-1)\n");
                uPackageGoalNodes.append("      (at uPackage-" + pack + " uNode-" + (((pack - 1) % (numberOfNodesUnsolvable - 1)) + 2) + ")\n");
            }
        }
        
        StringBuilder sPackageStartNodes = new StringBuilder(33 * numberOfPackagesSolvable); // capacity estimated
        StringBuilder sPackageGoalNodes = new StringBuilder(33 * numberOfPackagesSolvable); // capacity estimated
        for (int pack = 1; pack <= numberOfPackagesSolvable; pack++) {
            int startNode = (int) (Math.random() * numberOfNodesSolvable) + 1;
            int goalNode = (int) (Math.random() * (numberOfNodesSolvable - 1)) + 1;
            if (goalNode >= startNode) {
                goalNode++;
            }
            sPackageStartNodes.append("    (at sPackage-" + pack + " sNode-" + startNode + ")\n");
            sPackageGoalNodes.append("      (at sPackage-" + pack + " sNode-" + goalNode + ")\n");
        }

        pddl.append(uPackageStartNodes);
        pddl.append(sPackageStartNodes);
        pddl.append("  )\n"); // init close
        pddl.append("  (:goal\n");
        pddl.append("    (and\n");
        pddl.append(uPackageGoalNodes);
        pddl.append(sPackageGoalNodes);
        pddl.append("  ))\n"); // goal+and close
        pddl.append(")\n"); // problem close
        return pddl.toString();
    }

    /**
     * Returns the Name that should be used for the generated problem PDDL
     * file(s). The files will look like this: < br >
     * [getProblemFileName][problem-file-suffix][n][pddl-file-ending] < br >
     * e.g. nomystery-problem1.pddl nomystery-problem2.pddl ...< br >
     * If [getProblemFileName] is null, the name of the problem javafile will be
     * used. [n] is the number (importent if you create several instances in one
     * go). [problem-file-suffix] and [pddl-file-ending] are specified in the
     * properties file. Please beware of the naming restrictions of your system.
     *
     * @return the problem-name to be used to name the output files or null
     */
    @Override
    public String getProblemFileName() {
        return "nomystery-";
    }

    /**
     * Returns the Name that should be used for the generated domain PDDL file.
     * The file will look like this: < br >
     * [getDomainFileName][problem-file-suffix][pddl-file-ending] < br >
     * e.g. nomystery-domain.pddl< br >
     * If [getDomainFileName] is null, the name of the problem javafile will be
     * used. [domain-file-suffix] and [pddl-file-ending] are specified in the
     * properties file. Please beware of the naming restrictions of your system.
     *
     * @return the problem-name to be used to name the output files or null
     */
    @Override
    public String getDomainFileName() {
        return "";
    }
}
