
/**
 * A PDDL generator that generates PDDL descriptions of domains and problems of
 * Sokoban.
 *
 * @author Dietrich Zerr
 */
public class Sokoban implements Problem {

    /**
     * Unsolvability mode = "What kind of unsolvability do you want?".< br >
     * See Readme for details.
     */
    int mode;
    /**
     * Horizontal field size.
     */
    int xSize = 10;
    /**
     * Vertical field size.
     */
    int ySize = 10;
    /**
     * Number of stones on the field (unsually = number of goals). The mode's or
     * pattern's fixed stones are included in this number.
     */
    int stones = 5;

    /**
     * The domain of a Sokoban instance can be the same for all problems. It's a
     * static domain PDDL taken from the Fast Downward benchmarks folder.
     *
     * @param properties - not used
     * @return a Sokoban PDDL domain description
     */
    @Override
    public String getPDDLDomain(int[] properties) throws IllegalArgumentException {
        if (properties.length == 0) {
            throw new IllegalArgumentException("Please specify at least the unsolvability mode for Sokoban.\n"
                    + "The Sokoban call should look like this:\n"
                    + "Problemgenerator [<#problems to generate>] Sokoban <mode> [<x-size>] [<y-size>] [<number of stones>]\n"
                    + "Please see the readme for more details.");
        }

        StringBuilder pddl = new StringBuilder(2400);
        pddl.append("(define (domain sokoban)\n");
        pddl.append("  (:requirements :typing :action-costs)\n");
        pddl.append("  (:types thing location direction - object\n");
        pddl.append("          player stone - thing)\n");
        pddl.append("  (:predicates (clear ?l - location)\n");
        pddl.append("	       (at ?t - thing ?l - location)\n");
        pddl.append("	       (at-goal ?s - stone)\n");
        pddl.append("	       (IS-GOAL ?l - location)\n");
        pddl.append("	       (IS-NONGOAL ?l - location)\n");
        pddl.append("               (MOVE-DIR ?from ?to - location ?dir - direction))\n");
        pddl.append("  (:functions (total-cost) - number)\n");
        pddl.append("\n");
        pddl.append("  (:action move\n");
        pddl.append("   :parameters (?p - player ?from ?to - location ?dir - direction)\n");
        pddl.append("   :precondition (and (at ?p ?from)\n");
        pddl.append("                      (clear ?to)\n");
        pddl.append("                      (MOVE-DIR ?from ?to ?dir)\n");
        pddl.append("                      )\n");
        pddl.append("   :effect       (and (not (at ?p ?from))\n");
        pddl.append("                      (not (clear ?to))\n");
        pddl.append("                      (at ?p ?to)\n");
        pddl.append("                      (clear ?from)\n");
        pddl.append("                      )\n");
        pddl.append("   )\n");
        pddl.append("\n");
        pddl.append("  (:action push-to-nongoal\n");
        pddl.append("   :parameters (?p - player ?s - stone\n");
        pddl.append("                ?ppos ?from ?to - location\n");
        pddl.append("                ?dir - direction)\n");
        pddl.append("   :precondition (and (at ?p ?ppos)\n");
        pddl.append("                      (at ?s ?from)\n");
        pddl.append("                      (clear ?to)\n");
        pddl.append("                      (MOVE-DIR ?ppos ?from ?dir)\n");
        pddl.append("                      (MOVE-DIR ?from ?to ?dir)\n");
        pddl.append("                      (IS-NONGOAL ?to)\n");
        pddl.append("                      )\n");
        pddl.append("   :effect       (and (not (at ?p ?ppos))\n");
        pddl.append("                      (not (at ?s ?from))\n");
        pddl.append("                      (not (clear ?to))\n");
        pddl.append("                      (at ?p ?from)\n");
        pddl.append("                      (at ?s ?to)\n");
        pddl.append("                      (clear ?ppos)\n");
        pddl.append("                      (not (at-goal ?s))\n");
        pddl.append("                      (increase (total-cost) 1)\n");
        pddl.append("                      )\n");
        pddl.append("   )\n");
        pddl.append("\n");
        pddl.append("  (:action push-to-goal\n");
        pddl.append("   :parameters (?p - player ?s - stone\n");
        pddl.append("                ?ppos ?from ?to - location\n");
        pddl.append("                ?dir - direction)\n");
        pddl.append("   :precondition (and (at ?p ?ppos)\n");
        pddl.append("                      (at ?s ?from)\n");
        pddl.append("                      (clear ?to)\n");
        pddl.append("                      (MOVE-DIR ?ppos ?from ?dir)\n");
        pddl.append("                      (MOVE-DIR ?from ?to ?dir)\n");
        pddl.append("                      (IS-GOAL ?to)\n");
        pddl.append("                      )\n");
        pddl.append("   :effect       (and (not (at ?p ?ppos))\n");
        pddl.append("                      (not (at ?s ?from))\n");
        pddl.append("                      (not (clear ?to))\n");
        pddl.append("                      (at ?p ?from)\n");
        pddl.append("                      (at ?s ?to)\n");
        pddl.append("                      (clear ?ppos)\n");
        pddl.append("                      (at-goal ?s)\n");
        pddl.append("                      (increase (total-cost) 1)\n");
        pddl.append("                      )\n");
        pddl.append("   )\n");
        pddl.append(")");
        return pddl.toString();
    }

    /**
     * A Sokoban problem depends on the attributes. (x = 1, y = 1) is the field
     * in the top left corner. There is no 0 coordinate. < br >
     * attributes array: < br >
     * index - attribute < br >
     * 0 - unsolvability mode < br >
     * 1 - x-size of the field < br >
     * 2 - y-size of the field < br >
     * 3 - number of stones < br >
     *
     * @param properties Sokoban problem properties
     * @param instanceNumber the number of the problem instance
     * @return a Sokoban PDDL problem description
     */
    @Override
    public String getPDDLProblem(int[] properties, int instanceNumber) throws IllegalArgumentException {
        mode = properties[0];
        if (properties.length >= 2) {
            xSize = properties[1];
            if (xSize < 3) {
                xSize = 3;
                System.out.println("Horizontal field size can't be smaller then 3. Size has been set to 3.");
            }
        }
        if (properties.length >= 3) {
            ySize = properties[2];
            if (ySize < 3) {
                ySize = 3;
                System.out.println("Vertical field size can't be smaller then 3. Size has been set to 3.");
            }
        }
        if (properties.length >= 4) {
            stones = properties[3];
        }
        Positions positions = new Positions(instanceNumber); // this call generates everything
        StringBuilder pddl = new StringBuilder(230 * xSize * ySize); // capacity estimated

        pddl.append(";; Mode: " + mode + "\n");
        pddl.append(";; Field size: " + xSize + "x" + ySize + "\n");
        pddl.append(";; " + positions.stonesPlaced + " Stones\n\n");
        if (xSize <= 30 && ySize <= 30) {
            pddl.append(";; Legend:\n");
            pddl.append(";; # = wall\n");
            pddl.append(";; O = stone\n");
            pddl.append(";; @ = goal\n");
            pddl.append(";; P = player\n");
            pddl.append(positions);
            pddl.append("\n");
        }
        pddl.append("(define (problem sokoban-problem)\n");
        pddl.append("  (:domain sokoban)\n");
        pddl.append("  (:objects\n");
        pddl.append("    dir-down - direction\n");
        pddl.append("    dir-left - direction\n");
        pddl.append("    dir-right - direction\n");
        pddl.append("    dir-up - direction\n");
        pddl.append("    player-1 - player\n");
        for (int x = 1; x <= xSize; x++) {
            for (int y = 1; y <= ySize; y++) {
                pddl.append("    pos-" + x + "-" + y + " - location\n");;
            }
        }
        for (int s = 1; s <= positions.stonesPlaced; s++) {
            pddl.append("    stone-" + s + " - stone\n");
        }
        pddl.append("  )\n"); // objects close
        pddl.append("  (:init\n");
        for (int x = 2; x < xSize; x++) {
            for (int y = 2; y < ySize; y++) {
                pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x - 1) + "-" + y + " dir-left)\n");
                pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y - 1) + " dir-up)\n");
                pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x + 1) + "-" + y + " dir-right)\n");
                pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y + 1) + " dir-down)\n");
            }
        }
        for (int x = 2; x < xSize; x++) {
            int y = 1;
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x - 1) + "-" + y + " dir-left)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x + 1) + "-" + y + " dir-right)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y + 1) + " dir-down)\n");
            y = ySize;
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x - 1) + "-" + y + " dir-left)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y - 1) + " dir-up)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x + 1) + "-" + y + " dir-right)\n");
        }
        for (int y = 2; y < ySize; y++) {
            int x = 1;
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y - 1) + " dir-up)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x + 1) + "-" + y + " dir-right)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y + 1) + " dir-down)\n");
            x = xSize;
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + (x - 1) + "-" + y + " dir-left)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y - 1) + " dir-up)\n");
            pddl.append("    (MOVE-DIR pos-" + x + "-" + y + " pos-" + x + "-" + (y + 1) + " dir-down)\n");
        }
        pddl.append("    (MOVE-DIR pos-" + 1 + "-" + 1 + " pos-" + 1 + "-" + 2 + " dir-down)\n");
        pddl.append("    (MOVE-DIR pos-" + 1 + "-" + 1 + " pos-" + 2 + "-" + 1 + " dir-right)\n");
        pddl.append("    (MOVE-DIR pos-" + 1 + "-" + ySize + " pos-" + 2 + "-" + ySize + " dir-right)\n");
        pddl.append("    (MOVE-DIR pos-" + 1 + "-" + ySize + " pos-" + 1 + "-" + (ySize - 1) + " dir-up)\n");
        pddl.append("    (MOVE-DIR pos-" + xSize + "-" + 1 + " pos-" + (xSize - 1) + "-" + 1 + " dir-left)\n");
        pddl.append("    (MOVE-DIR pos-" + xSize + "-" + 1 + " pos-" + xSize + "-" + 2 + " dir-down)\n");
        pddl.append("    (MOVE-DIR pos-" + xSize + "-" + ySize + " pos-" + (xSize - 1) + "-" + ySize + " dir-left)\n");
        pddl.append("    (MOVE-DIR pos-" + xSize + "-" + ySize + " pos-" + xSize + "-" + (ySize - 1) + " dir-up)\n");

        int stoneCount = 1;
        for (int y = 1; y <= ySize; y++) {
            for (int x = 1; x <= xSize; x++) {
                switch (positions.field[x - 1][y - 1]) {
                    case -1:
                        pddl.append("    (IS-NONGOAL pos-" + x + "-" + y + ")\n");
                        break;
                    case 0:
                        pddl.append("    (IS-NONGOAL pos-" + x + "-" + y + ")\n");
                        pddl.append("    (clear pos-" + x + "-" + y + ")\n");
                        break;
                    case 1:
                        pddl.append("    (IS-NONGOAL pos-" + x + "-" + y + ")\n");
                        pddl.append("    (at stone-" + stoneCount + " pos-" + x + "-" + y + ")\n");
                        stoneCount++;
                        break;
                    case 2:
                        pddl.append("    (IS-GOAL pos-" + x + "-" + y + ")\n");
                        pddl.append("    (clear pos-" + x + "-" + y + ")\n");
                        break;
                    case 3:
                        pddl.append("    (at player-1 pos-" + x + "-" + y + ")\n");
                        pddl.append("    (IS-NONGOAL pos-" + x + "-" + y + ")\n");
                }
            }
        }

        pddl.append("    (= (total-cost) 0)\n");
        pddl.append("  )\n"); // init close
        pddl.append("  (:goal (and\n");
        for (int s = 1; s <= positions.stonesPlaced; s++) {
            pddl.append("    (at-goal stone-" + s + ")\n");
        }
        pddl.append("  ))\n"); // goal close
        pddl.append("  (:metric minimize (total-cost))\n");
        pddl.append(")");

        return pddl.toString();
    }

    /**
     * Returns the Name that should be used for the generated problem PDDL
     * file(s). The files will look like this: < br >
     * [getProblemFileName][problem-file-suffix][n][pddl-file-ending] < br >
     * e.g. sokoban-problem1.pddl sokoban-problem2.pddl ...< br >
     * If [getProblemFileName] is null, the name of the problem javafile will be
     * used. [n] is the number (importent if you create several instances in one
     * go). [problem-file-suffix] and [pddl-file-ending] are specified in the
     * properties file. Please beware of the naming restrictions of your system.
     *
     * @return the problem-name to be used to name the output files or null
     */
    @Override
    public String getProblemFileName() {
        return "sokoban-";
    }

    /**
     * Returns the Name that should be used for the generated domain PDDL file.
     * The file will look like this: < br >
     * [getDomainFileName][problem-file-suffix][pddl-file-ending] < br >
     * e.g. sokoban-domain.pddl< br >
     * If [getDomainFileName] is null, the name of the problem javafile will be
     * used. [domain-file-suffix] and [pddl-file-ending] are specified in the
     * properties file. Please beware of the naming restrictions of your system.
     *
     * @return the problem-name to be used to name the output files or null
     */
    @Override
    public String getDomainFileName() {
        return "";
    }

    /**
     * This class has the logic to decide where to put all objects on the field
     * in order to create the desired unsolvable instance, but it's also a
     * structure to store all those positions. Unspecific stones are spread
     * randomly across the field but not on the field's edges and not next to
     * each other. Goal fields may be on edges and next to each other.
     */
    private class Positions {

        /**
         * The sokoban field represented in integer values. < br >
         * -1 = wall/not clear < br >
         * 0 = clear < br >
         * 1 = stone < br >
         * 2 = goal-field
         */
        public int field[][] = new int[xSize][ySize];
        public int stonesPlaced = 0;
        public int goalsPlaced = 0;
        public int positioningRetries = 1000; // how often is random positioning allowed to fail/retry

        /**
         * The constructor sets all positions.
         *
         * @param instanceNumber the number of the current instance
         */
        public Positions(int instanceNumber) throws IllegalArgumentException {
            int stones_originally = stones;
            switch (mode) {
                case 0: // = one stone is on the upper edge of the field
                    //     1 2 3 4 5
                    //   # # # # # # #
                    // 1 # P   O     #
                    // 2 #           #
                    // 3 #       @   #
                    // 4 #           #
                    //   # # # # # # #
                    // placeholders to keep the uppter edge clear
                    validate(3, 3, 1, instanceNumber);
                    for (int x = 0; x < xSize; x++) {
                        field[x][0] = -2;
                    }
                    stones--; // the one stone that we put on the upper edge later
                    placeObjects(instanceNumber);
                    for (int x = 0; x < xSize; x++) {
                        field[x][0] = 0;
                    }
                    // now placing the last stone and goal
                    field[xSize / 2][0] = 1;
                    stonesPlaced++;
                    while (true) {
                        int x = (int) (Math.random() * xSize);
                        int y = (int) (Math.random() * (ySize - 1) + 1);
                        if (field[x][y] == 0) {
                            field[x][y] = 2;
                            goalsPlaced++;
                            break;
                        }
                    }
                    break;

                case 1: // "pattern" 0a
                    //     1 2 3 4 5 6
                    //   # # # # # # # #
                    // 1 # P           #
                    // 2 #         # # #
                    // 3 #       #     #
                    // 4 #       #   O #
                    // 5 #       #   @ #
                    //   # # # # # # # #
                    validate(6, 5, 1, instanceNumber);
                    // reserve space
                    for (int y = ySize - 4; y < ySize; y++) {
                        for (int x = xSize - 3; x < xSize; x++) {
                            field[x][y] = -1;
                        }
                    }
                    stones--;
                    placeObjects(instanceNumber);
                    for (int y = ySize - 4; y < ySize; y++) {
                        for (int x = xSize - 3; x < xSize; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 1][ySize - 2] = 1;
                    field[xSize - 1][ySize - 1] = 2;
                    stonesPlaced++;
                    goalsPlaced++;
                    field[xSize - 3][ySize - 1] = -1;
                    field[xSize - 3][ySize - 2] = -1;
                    field[xSize - 3][ySize - 3] = -1;
                    field[xSize - 2][ySize - 4] = -1;
                    field[xSize - 1][ySize - 4] = -1;
                    break;

                case 2: // not enough goals
                    field[1][1] = 1;
                    stones--;
                    placeObjects(instanceNumber);
                    stonesPlaced++;
                    break;

                case 3: // "pattern" 0b
                    //     1 2 3 4 5 6
                    //   # # # # # # # #
                    // 1 # P           #
                    // 2 #         # # #
                    // 3 #       #     #
                    // 4 #   O   #     #
                    // 5 #       #   @ #
                    //   # # # # # # # #
                    validate(6, 5, 1, instanceNumber);
                    // reserve space
                    for (int y = ySize - 4; y < ySize; y++) {
                        for (int x = xSize - 3; x < xSize; x++) {
                            field[x][y] = -1;
                        }
                    }
                    field[xSize - 5][ySize - 2] = 1;
                    stones--;
                    placeObjects(instanceNumber);
                    stonesPlaced++;
                    for (int y = ySize - 4; y < ySize; y++) {
                        for (int x = xSize - 3; x < xSize; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 1][ySize - 1] = 2;
                    goalsPlaced++;
                    field[xSize - 3][ySize - 1] = -1;
                    field[xSize - 3][ySize - 2] = -1;
                    field[xSize - 3][ySize - 3] = -1;
                    field[xSize - 2][ySize - 4] = -1;
                    field[xSize - 1][ySize - 4] = -1;
                    break;

                case 4: // pattern 1
                    //     1 2 3 4 5
                    //   # # # # # # #
                    // 1 # P         #
                    // 2 #           #
                    // 3 #       # # #
                    // 4 #     O     #
                    // 5 #       # @ #
                    //   # # # # # # #
                    validate(5, 5, 1, instanceNumber);
                    // reserve space
                    for (int y = ySize - 3; y < ySize; y++) {
                        for (int x = xSize - 2; x < xSize; x++) {
                            field[x][y] = -2;
                        }
                    }
                    field[xSize - 3][ySize - 2] = 1;
                    stones--;
                    placeObjects(instanceNumber);
                    stonesPlaced++;
                    for (int y = ySize - 3; y < ySize; y++) {
                        for (int x = xSize - 2; x < xSize; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 1][ySize - 1] = 2;
                    goalsPlaced++;
                    field[xSize - 2][ySize - 1] = -1;
                    field[xSize - 2][ySize - 3] = -1;
                    field[xSize - 1][ySize - 3] = -1;
                    break;

                case 5: // pattern 2
                    //     1 2 3 4 5 6 7
                    //   # # # # # # # # #
                    // 1 # P             #
                    // 2 #       #       #
                    // 3 #     O         #
                    // 4 #   #   @   #   #
                    // 5 #               #
                    // 6 #       #       #
                    // 7 #               #
                    //   # # # # # # # # #
                    validate(7, 7, 1, instanceNumber);
                    for (int y = ySize - 6; y < ySize - 1; y++) {
                        for (int x = xSize - 6; x < xSize - 1; x++) {
                            field[x][y] = -1;
                        }
                    }
                    stones--;
                    placeObjects(instanceNumber);
                    stonesPlaced++;
                    for (int y = ySize - 6; y < ySize - 1; y++) {
                        for (int x = xSize - 6; x < xSize - 1; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 4][ySize - 2] = -1;
                    field[xSize - 4][ySize - 6] = -1;
                    field[xSize - 2][ySize - 4] = -1;
                    field[xSize - 6][ySize - 4] = -1;
                    field[xSize - 4][ySize - 4] = 2;
                    goalsPlaced++;
                    field[xSize - 5][ySize - 5] = 1;
                    break;

                case 6: // pattern 3
                    //     1 2 3 4 5 6 7 8
                    //   # # # # # # # # # #
                    // 1 # P               #
                    // 2 #                 #
                    // 3 #       # # #     #
                    // 4 #       #   #     #
                    // 5 #                 #
                    // 6 #     # O @   #   #
                    // 7 #                 #
                    // 8 #       #   #     #
                    //   # # # # # # # # # #
                    validate(8, 8, 1, instanceNumber);
                    for (int y = ySize - 6; y < ySize; y++) {
                        for (int x = xSize - 5; x < xSize - 2; x++) {
                            field[x][y] = -1;
                        }
                    }
                    field[xSize - 2][ySize - 3] = -1;
                    field[xSize - 6][ySize - 3] = -1;
                    field[xSize - 2][ySize - 1] = -1;
                    field[xSize - 2][ySize - 2] = -1;
                    field[xSize - 2][ySize - 4] = -1;
                    stones--;
                    placeObjects(instanceNumber);
                    stonesPlaced++;
                    for (int y = ySize - 4; y < ySize - 1; y++) {
                        for (int x = xSize - 5; x < xSize - 2; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 4][ySize - 5] = 0;
                    field[xSize - 4][ySize - 1] = 0;
                    field[xSize - 2][ySize - 1] = 0;
                    field[xSize - 2][ySize - 2] = 0;
                    field[xSize - 2][ySize - 4] = 0;
                    field[xSize - 4][ySize - 3] = 2;
                    goalsPlaced++;
                    field[xSize - 5][ySize - 3] = 1;
                    break;

                case 7: // pattern 4
                    //     1 2 3 4 5
                    //   # # # # # # #
                    // 1 # P         #
                    // 2 #           #
                    // 3 #     @ O # #
                    // 4 #     O @   #
                    // 5 #     #   # #
                    //   # # # # # # #
                    validate(5, 5, 2, instanceNumber);
                    field[xSize - 1][ySize - 1] = -1;
                    field[xSize - 2][ySize - 1] = -2;
                    field[xSize - 3][ySize - 1] = -1;
                    field[xSize - 1][ySize - 2] = -2;
                    field[xSize - 2][ySize - 2] = 2;
                    field[xSize - 3][ySize - 2] = 1;
                    field[xSize - 1][ySize - 3] = -1;
                    field[xSize - 2][ySize - 3] = 1;
                    field[xSize - 3][ySize - 3] = 2;
                    stones -= 2;
                    placeObjects(instanceNumber);
                    stonesPlaced += 2;
                    goalsPlaced += 2;
                    field[xSize - 2][ySize - 1] = 0;
                    field[xSize - 1][ySize - 2] = 0;
                    break;

                case 8: // pattern 5
                    //     1 2 3 4 5
                    //   # # # # # # #
                    // 1 # P         #
                    // 2 #       @ @ #
                    // 3 #     # O O #
                    // 4 #     #     #
                    // 5 #     #     #
                    //   # # # # # # #
                    validate(5, 5, 2, instanceNumber);
                    for (int y = ySize - 2; y < ySize; y++) {
                        for (int x = xSize - 3; x < xSize; x++) {
                            field[x][y] = -1;
                        }
                    }
                    field[xSize - 3][ySize - 3] = -1;
                    field[xSize - 2][ySize - 3] = 1;
                    field[xSize - 2][ySize - 4] = 2;
                    field[xSize - 1][ySize - 3] = 1;
                    field[xSize - 1][ySize - 4] = 2;
                    stones -= 2;
                    placeObjects(instanceNumber);
                    stonesPlaced += 2;
                    goalsPlaced += 2;
                    for (int y = ySize - 2; y < ySize; y++) {
                        for (int x = xSize - 2; x < xSize; x++) {
                            field[x][y] = 0;
                        }
                    }
                    break;

                case 9: // pattern 6
                    //     1 2 3 4 5 6 7 8 9 0 1
                    //   # # # # # # # # # # # # #
                    // 1 # P                     #
                    // 2 #                       #
                    // 3 #     # # # # # # # #   #
                    // 4 #     #             #   #
                    // 5 #         @ O O @       #
                    // 6 #     #             #   #
                    //   # # # # # # # # # # # # #
                    validate(11, 6, 2, instanceNumber);
                    for (int y = ySize - 4; y < ySize; y++) {
                        for (int x = xSize - 9; x < xSize - 1; x++) {
                            field[x][y] = -1;
                        }
                    }
                    stones -= 2;
                    placeObjects(instanceNumber);
                    stonesPlaced += 2;
                    for (int y = ySize - 3; y < ySize; y++) {
                        for (int x = xSize - 8; x < xSize - 2; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 2][ySize - 2] = 0;
                    field[xSize - 9][ySize - 2] = 0;
                    field[xSize - 4][ySize - 2] = 2;
                    field[xSize - 5][ySize - 2] = 1;
                    field[xSize - 6][ySize - 2] = 1;
                    field[xSize - 7][ySize - 2] = 2;
                    goalsPlaced += 2;
                    break;

                case 10: // pattern 7
                    //     1 2 3 4 5 6 7 8 9 10
                    //   # # # # # # # # # # # #
                    // 1 # P                   #
                    // 2 #                     #
                    // 3 #     # # # #   # # # #
                    // 4 #     #   @       @   #
                    // 5 #     #   O #   # O   #
                    // 6 #     #     #   #     #
                    // 7 #     # # # #   # # # #
                    // 8 #                     #
                    // 9 #                     #
                    //   # # # # # # # # # # # #
                    validate(10, 9, 2, instanceNumber);
                    for (int y = ySize - 7; y < ySize - 2; y++) {
                        for (int x = xSize - 8; x < xSize; x++) {
                            field[x][y] = -1;
                        }
                    }
                    field[xSize - 4][ySize - 2] = 1;
                    field[xSize - 4][ySize - 8] = 1;
                    stones -= 2;
                    placeObjects(instanceNumber);
                    stonesPlaced += 2;
                    for (int y = ySize - 6; y < ySize - 3; y++) {
                        for (int x = xSize - 7; x < xSize; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 4][ySize - 2] = 0;
                    field[xSize - 4][ySize - 8] = 0;
                    field[xSize - 4][ySize - 3] = 0;
                    field[xSize - 4][ySize - 7] = 0;
                    field[xSize - 2][ySize - 5] = 1;
                    field[xSize - 6][ySize - 5] = 1;
                    field[xSize - 2][ySize - 6] = 2;
                    field[xSize - 6][ySize - 6] = 2;
                    goalsPlaced += 2;
                    field[xSize - 3][ySize - 5] = -1;
                    field[xSize - 5][ySize - 5] = -1;
                    field[xSize - 3][ySize - 4] = -1;
                    field[xSize - 5][ySize - 4] = -1;
                    break;

                case 11: // pattern 8
                    //     1 2 3 4 5 6 7 8 9 10
                    //   # # # # # # # # # # # #
                    // 1 #                     #
                    // 2 #     O # # # # # O   #
                    // 3 #       @   #   @     #
                    // 4 #     # # @ # @ # #   #
                    // 5 #     #           #   #
                    // 6 #     #   O   O   #   #
                    // 7 #     #           #   #
                    //   # # # # # # # # # # # #
                    validate(10, 7, 4, instanceNumber);
                    for (int y = ySize - 6; y < ySize; y++) {
                        for (int x = xSize - 8; x < xSize - 1; x++) {
                            field[x][y] = -1;
                        }
                    }
                    field[xSize - 2][ySize - 6] = 1;
                    field[xSize - 8][ySize - 6] = 1;
                    stones -= 4;
                    placeObjects(instanceNumber);
                    for (int y = ySize - 3; y < ySize; y++) {
                        for (int x = xSize - 7; x < xSize - 2; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 4][ySize - 2] = 1;
                    field[xSize - 6][ySize - 2] = 1;
                    stonesPlaced += 4;
                    field[xSize - 4][ySize - 4] = 2;
                    field[xSize - 6][ySize - 4] = 2;
                    field[xSize - 7][ySize - 5] = 2;
                    field[xSize - 3][ySize - 5] = 2;
                    goalsPlaced += 4;
                    field[xSize - 8][ySize - 5] = 0;
                    field[xSize - 6][ySize - 5] = 0;
                    field[xSize - 4][ySize - 5] = 0;
                    field[xSize - 2][ySize - 5] = 0;
                    break;

                case 12: // pattern 9
                    //     1 2 3 4 5 6 7 8 
                    //   # # # # # # # # # #
                    // 1 #                 #
                    // 2 #                 #
                    // 3 #     O   # # # # #
                    // 4 #   O @   #   @ @ #
                    // 5 #     #       O   #
                    // 6 #         #       #
                    //   # # # # # # # # # #
                    validate(8, 6, 3, instanceNumber);
                    for (int y = ySize - 4; y < ySize; y++) {
                        for (int x = xSize - 7; x < xSize; x++) {
                            field[x][y] = -1;
                        }
                    }
                    field[xSize - 6][ySize - 4] = 1;
                    field[xSize - 7][ySize - 3] = 1;
                    stones -= 3;
                    placeObjects(instanceNumber);
                    for (int y = ySize - 4; y < ySize; y++) {
                        for (int x = xSize - 7; x < xSize; x++) {
                            field[x][y] = 0;
                        }
                    }
                    field[xSize - 6][ySize - 4] = 1;
                    field[xSize - 7][ySize - 3] = 1;
                    field[xSize - 2][ySize - 2] = 1;
                    stonesPlaced += 3;
                    field[xSize - 1][ySize - 3] = 2;
                    field[xSize - 2][ySize - 3] = 2;
                    field[xSize - 6][ySize - 3] = 2;
                    goalsPlaced += 3;
                    field[xSize - 1][ySize - 4] = -1;
                    field[xSize - 2][ySize - 4] = -1;
                    field[xSize - 3][ySize - 4] = -1;
                    field[xSize - 4][ySize - 4] = -1;
                    field[xSize - 4][ySize - 3] = -1;
                    field[xSize - 4][ySize - 1] = -1;
                    field[xSize - 6][ySize - 2] = -1;
                    break;

                default:
                    throw new IllegalArgumentException("Invalid mode: " + mode + "\nSee the readme to see which modes are available.");
            }
            stones = stones_originally;
        }

        /**
         * This method validates the parameters and corrects them if necessary.
         *
         * @param minX the required horizontal size of the field
         * @param minY the required vertical size of the field
         * @param minStones the required amount of stones
         */
        private void validate(int minX, int minY, int minStones, int instanceNumber) {
            if (xSize < minX) {
                xSize = minX;
                if (instanceNumber == 1) { 
                    System.out.println("Mode " + mode + " requires horizontal size to be at least " + minX + ". Size set to " + minX + ".");
                }
                field = new int[xSize][ySize];
            }
            if (ySize < minY) {
                ySize = minY;
                if (instanceNumber == 1) { 
                    System.out.println("Mode " + mode + " requires vertical size to be at least " + minY + ". Size set to " + minY + ".");
                }
                field = new int[xSize][ySize];
            }
            if (stones < minStones) {
                stones = minStones;
                if (instanceNumber == 1) { 
                    System.out.println("Mode " + mode + " requires at least " + minStones + " stones. Stones set to " + minStones + ".");
                }
            }
        }

        /**
         * This method places as many random stones and goals as the current
         * value in stones. Stones are never placed next to the outer edges or
         * next to each other. Goals can be on any empty field. The player is
         * always placed on (1,1).
         *
         * @param instanceNumber the number of the current instance
         */
        private void placeObjects(int instanceNumber) {
            field[0][0] = 3;
            int x;
            int y;
            for (int i = 0; i < stones; i++) {
                for (int tries = 0; tries < positioningRetries; tries++) {
                    x = (int) (Math.random() * (xSize - 2) + 1);
                    y = (int) (Math.random() * (ySize - 2) + 1);
                    if (field[x][y] == 0
                            && field[x - 1][y] != 1
                            && field[x][y - 1] != 1
                            && field[x + 1][y] != 1
                            && field[x][y + 1] != 1
                            && field[x - 1][y - 1] != 1
                            && field[x - 1][y + 1] != 1
                            && field[x + 1][y - 1] != 1
                            && field[x + 1][y + 1] != 1) {
                        field[x][y] = 1;
                        stonesPlaced++;
                        break;
                    }
                }
            }
            while (stonesPlaced > goalsPlaced) {
                x = (int) (Math.random() * xSize);
                y = (int) (Math.random() * ySize);
                if (field[x][y] == 0) {
                    field[x][y] = 2;
                    goalsPlaced++;
                }
            }
            if (stones > stonesPlaced) {
                System.out.println("[Instance " + instanceNumber + "] Warning: Only " + stonesPlaced + " random stones could be placed instead of " + stones + ".");
            }
        }

        @Override
        public String toString() {
            String output = "";
            String columnCount = ";;    ";
            for (int x = 1; x <= xSize; x++) {
                columnCount += " " + x % 10;
            }
            columnCount += " \n";
            output += columnCount;
            String borderline = ";;   #";
            for (int x = 0; x < xSize; x++) {
                borderline += " #";
            }
            borderline += " #\n";
            output += borderline;
            for (int y = 0; y < ySize; y++) {
                output += ";; " + (y + 1) % 10 + " #";
                for (int x = 0; x < xSize; x++) {
                    output += " ";
                    switch (field[x][y]) {
                        case -1:
                            output += "#";
                            break;
                        case 0:
                            output += " ";
                            break;
                        case 1:
                            output += "O";
                            break;
                        case 2:
                            output += "@";
                            break;
                        case 3:
                            output += "P";
                            break;
                        default:
                            System.out.println("Found a " + field[x][y] + " on field[" + x + "][" + y + "]");
                    }
                }
                output += " #\n";
            }
            return output + borderline;
        }
    }
}
